#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Monkey patch Requests/Urllib3 to allow restricting the address family"""

import threading
from requests.packages.urllib3.util import connection

_allowed_gai_family_orig = connection.allowed_gai_family

_allowed_family_mutex = threading.RLock()
_allowed_family = None


def _allowed_gai_family():
    with _allowed_family_mutex:
        if _allowed_family is None:
            return _allowed_gai_family_orig()
        else:
            return _allowed_family


connection.allowed_gai_family = _allowed_gai_family


class RequestsFamilyRestriction:
    """Context manager that causes Requests to only use the specified address
    family.

    For example, to force a request over IPv6::

        with RequestsFamilyRestriction(socket.AF_INET6):
            r = requests.get("https://icanhazip.com/")
        # error checking skipped
        ipv6 = r.text

    :param family: The address family to use (typically a ``socket.AF_*``
                   constant from Python's :mod:`socket` module)
    """

    def __init__(self, family):
        self.family = family

    def __enter__(self):
        global _allowed_family
        _allowed_family_mutex.acquire()
        _allowed_family = self.family

    def __exit__(self, exc_type, exc_val, exc_tb):
        global _allowed_family
        _allowed_family = None
        _allowed_family_mutex.release()



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import ipaddress

import ruddr


def test_require_and_skip_ipv4(notifier_factory):
    """Test requiring IPv4 when skipped is error"""
    with pytest.raises(ruddr.ConfigError):
        notifier_factory(ipv4_required='true', skip_ipv4='true')


def test_require_and_skip_ipv6(notifier_factory):
    """Test requiring IPv6 when skipped is error"""
    with pytest.raises(ruddr.ConfigError):
        notifier_factory(ipv6_required='true', skip_ipv6='true')


def test_require_ipv4_and_skip_ipv6(notifier_factory, mock_updater):
    """Test requiring IPv4 when IPv6 skipped is okay"""
    fake_notifier = notifier_factory(ipv4_required='true', skip_ipv6='true')
    fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)

    fake_notifier.notify_ipv4(ipaddress.IPv4Address('10.20.30.40'))
    assert mock_updater.published_addresses == [
        ipaddress.IPv4Address('10.20.30.40'),
    ]


def test_require_ipv6_and_skip_ipv4(notifier_factory, mock_updater):
    """Test requiring IPv6 when IPv4 skipped is okay"""
    fake_notifier = notifier_factory(ipv6_required='true', skip_ipv4='true')
    fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)

    fake_notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
    assert mock_updater.published_addresses == [
        ipaddress.IPv6Network('1234::/64'),
    ]


def test_ipv4_ready_false(notifier_factory, mock_updater):
    """Test attaching updaters with ipv4_ready set to false is error"""
    fake_notifier = notifier_factory(ipv4_ready='false')

    # Expect no exception for IPv6
    fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)
    with pytest.raises(ruddr.ConfigError):
        fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)


def test_ipv6_ready_false(notifier_factory, mock_updater):
    """Test attaching updaters with ipv6_ready set to false is error"""
    fake_notifier = notifier_factory(ipv6_ready='false')

    # Expect no exception for IPv4
    fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)
    with pytest.raises(ruddr.ConfigError):
        fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)


def test_ipv4_ready_false_attach_ipv6(notifier_factory, mock_updater):
    """Test attaching only IPv6 updater with ipv4_ready set to false"""
    fake_notifier = notifier_factory(ipv4_ready='false')
    fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('5678::/64'))
    # Well-written notifiers don't call notify_ipv4() when want_ipv4() is False

    assert mock_updater.published_addresses == [
        ipaddress.IPv6Network('1234::/64'),
        ipaddress.IPv6Network('5678::/64'),
    ]


def test_ipv6_ready_false_attach_ipv4(notifier_factory, mock_updater):
    """Test attaching only IPv4 updater with ipv6_ready set to false"""
    fake_notifier = notifier_factory(ipv6_ready='false')
    fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('10.20.30.40'))
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('50.60.70.80'))
    # Well-written notifiers don't call notify_ipv6() when want_ipv6() is False

    assert mock_updater.published_addresses == [
        ipaddress.IPv4Address('10.20.30.40'),
        ipaddress.IPv4Address('50.60.70.80'),
    ]


def test_want_ipv4_false(notifier_factory, mock_updater):
    """Test want_ipv4 is true when ipv4 update function not attached"""
    fake_notifier = notifier_factory()
    assert not fake_notifier.want_ipv4()


def test_want_ipv4_true(notifier_factory, mock_updater):
    """Test want_ipv4 is true when ipv4 update function attached"""
    fake_notifier = notifier_factory()
    fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)
    assert fake_notifier.want_ipv4()


def test_want_ipv6_false(notifier_factory, mock_updater):
    """Test want_ipv6 is true when ipv6 update function not attached"""
    fake_notifier = notifier_factory()
    assert not fake_notifier.want_ipv6()


def test_want_ipv6_true(notifier_factory, mock_updater):
    """Test want_ipv6 is true when ipv6 update function attached"""
    fake_notifier = notifier_factory()
    fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)
    assert fake_notifier.want_ipv6()


def test_need_ipv4_false_not_required(notifier_factory, mock_updater):
    """Test need_ipv4 is false when updater attached but not required"""
    fake_notifier = notifier_factory(ipv4_required='false')
    fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)
    assert not fake_notifier.need_ipv4()


def test_need_ipv4_false_no_updater(notifier_factory, mock_updater):
    """Test need_ipv4 is false when required but no updater attached"""
    fake_notifier = notifier_factory(ipv4_required='true')
    assert not fake_notifier.need_ipv4()


def test_need_ipv4_true(notifier_factory, mock_updater):
    """Test need_ipv4 is true when required and updater attached"""
    fake_notifier = notifier_factory(ipv4_required='true')
    fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)
    assert fake_notifier.need_ipv4()


def test_need_ipv6_false_not_required(notifier_factory, mock_updater):
    """Test need_ipv6 is false when updater attached but not required"""
    fake_notifier = notifier_factory(ipv6_required='false')
    fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)
    assert not fake_notifier.need_ipv6()


def test_need_ipv6_false_no_updater(notifier_factory, mock_updater):
    """Test need_ipv6 is false when required but no updater attached"""
    fake_notifier = notifier_factory(ipv6_required='true')
    assert not fake_notifier.need_ipv6()


def test_need_ipv6_true(notifier_factory, mock_updater):
    """Test need_ipv6 is true when required and updater attached"""
    fake_notifier = notifier_factory(ipv6_required='true')
    fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)
    assert fake_notifier.need_ipv6()


def test_defaults(notifier_factory, mock_updater):
    """Test default config settings are IPv4 required, IPv6 not, neither
    skipped"""
    fake_notifier = notifier_factory()
    fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)
    fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)
    assert fake_notifier.want_ipv4()
    assert fake_notifier.need_ipv4()
    assert fake_notifier.want_ipv6()
    assert not fake_notifier.need_ipv6()


def test_skip_ipv4_attach_ipv4(notifier_factory, mock_updater):
    """Test attaching an updater for IPv4 and notifying when IPv4 is skipped
    does nothing"""
    fake_notifier = notifier_factory(skip_ipv4='true')
    attached4 = fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)
    assert not attached4
    assert not fake_notifier.want_ipv4()
    # Well-behaved notifiers wouldn't notify_ipv4 when want_ipv4 is false, but
    # it must not cause problems
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('10.20.30.40'))
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('50.60.70.80'))
    assert mock_updater.published_addresses == []


def test_skip_ipv4_attach_both(notifier_factory, mock_updater):
    """Test attaching an updater when IPv4 is skipped works for IPv6"""
    fake_notifier = notifier_factory(skip_ipv4='true')
    attached4 = fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)
    attached6 = fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)
    assert not attached4
    assert attached6
    # Well-behaved notifiers wouldn't notify_ipv4 when want_ipv4 is false, but
    # it must not cause problems
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('10.20.30.40'))
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('50.60.70.80'))
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert mock_updater.published_addresses == [
        ipaddress.IPv6Network('1234::/64'),
        ipaddress.IPv6Network('5678::/64'),
    ]


def test_skip_ipv6_attach_ipv6(notifier_factory, mock_updater):
    """Test attaching an updater for IPv6 and notifying when IPv6 is skipped
    does nothing"""
    fake_notifier = notifier_factory(skip_ipv6='true')
    attached6 = fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)
    assert not attached6
    assert not fake_notifier.want_ipv6()
    # Well-behaved notifiers wouldn't notify_ipv6 when want_ipv6 is false, but
    # it must not cause problems
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('5678::/64'))
    assert mock_updater.published_addresses == []


def test_skip_ipv6_attach_both(notifier_factory, mock_updater):
    """Test attaching an updater when IPv6 is skipped works for IPv4"""
    fake_notifier = notifier_factory(skip_ipv6='true')
    attached4 = fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)
    attached6 = fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)
    assert attached4
    assert not attached6
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('10.20.30.40'))
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('50.60.70.80'))
    # Well-behaved notifiers wouldn't notify_ipv6 when want_ipv6 is false, but
    # it must not cause problems
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert mock_updater.published_addresses == [
        ipaddress.IPv4Address('10.20.30.40'),
        ipaddress.IPv4Address('50.60.70.80'),
    ]


def test_ipv4_ready_false_and_skipped(notifier_factory, mock_updater):
    """Test attaching updater with ipv4_ready set to false but IPv4 skipped
    does not raise error"""
    fake_notifier = notifier_factory(skip_ipv4='true', ipv4_ready='false')
    attached4 = fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)
    attached6 = fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)
    assert not attached4
    assert attached6
    # Well-behaved notifiers wouldn't notify_ipv4 when want_ipv4 is false
    # (especially when IPv4 isn't ready), but it must not cause problems
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('10.20.30.40'))
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('50.60.70.80'))
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert mock_updater.published_addresses == [
        ipaddress.IPv6Network('1234::/64'),
        ipaddress.IPv6Network('5678::/64'),
    ]


def test_ipv6_ready_false_and_skipped(notifier_factory, mock_updater):
    """Test attaching updater with ipv6_ready set to false but IPv6 skipped
    does not raise error"""
    fake_notifier = notifier_factory(skip_ipv6='true', ipv6_ready='false')
    attached4 = fake_notifier.attach_ipv4_updater(mock_updater.update_ipv4)
    attached6 = fake_notifier.attach_ipv6_updater(mock_updater.update_ipv6)
    assert attached4
    assert not attached6
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('10.20.30.40'))
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('50.60.70.80'))
    # Well-behaved notifiers wouldn't notify_ipv6 when want_ipv6 is false
    # (especially when IPv6 isn't ready), but it must not cause problems
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert mock_updater.published_addresses == [
        ipaddress.IPv4Address('10.20.30.40'),
        ipaddress.IPv4Address('50.60.70.80'),
    ]


def test_skip_both_error(notifier_factory):
    """Test that setting both skip_ipv4 and skip_ipv6 causes an exception"""
    with pytest.raises(ruddr.ConfigError):
        notifier_factory(skip_ipv4='true', skip_ipv6='true')


def test_multiple_updaters(notifier_factory, updater_factory):
    """Test multiple updaters attached to a single notifier"""
    fake_notifier = notifier_factory()
    mock_updater_1 = updater_factory()
    mock_updater_2 = updater_factory()
    fake_notifier.attach_ipv4_updater(mock_updater_1.update_ipv4)
    fake_notifier.attach_ipv6_updater(mock_updater_1.update_ipv6)
    fake_notifier.attach_ipv4_updater(mock_updater_2.update_ipv4)
    fake_notifier.attach_ipv6_updater(mock_updater_2.update_ipv6)

    fake_notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('10.20.30.40'))
    fake_notifier.notify_ipv4(ipaddress.IPv4Address('50.60.70.80'))
    fake_notifier.notify_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert mock_updater_1.published_addresses == [
        ipaddress.IPv6Network('1234::/64'),
        ipaddress.IPv4Address('10.20.30.40'),
        ipaddress.IPv4Address('50.60.70.80'),
        ipaddress.IPv6Network('5678::/64'),
    ]

    assert mock_updater_2.published_addresses == [
        ipaddress.IPv6Network('1234::/64'),
        ipaddress.IPv4Address('10.20.30.40'),
        ipaddress.IPv4Address('50.60.70.80'),
        ipaddress.IPv6Network('5678::/64'),
    ]



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Tests for Addrfile"""
import ipaddress

import pytest

import doubles
import ruddr.addrfile


@pytest.fixture
def addrfile_factory(tmp_path):
    count = 0

    def factory(contents: str):
        nonlocal count
        count += 1
        path = tmp_path / f"addrfile_{count}"

        with open(path, "w") as f:
            for line in contents.splitlines():
                print(line.strip(), file=f)

        return ruddr.addrfile.Addrfile(path)
    return factory


def test_new_addrfile(empty_addrfile):
    """Test new, nonexistent addrfile can get and set addresses"""
    assert empty_addrfile.get_ipv4("test") == (None, False)
    assert empty_addrfile.get_ipv6("test") == (None, False)
    assert empty_addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert empty_addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )

    empty_addrfile.set_ipv4("test", ipaddress.IPv4Address('1.2.3.4'))
    empty_addrfile.set_ipv6("test", ipaddress.IPv6Network('1234::/64'))
    assert (empty_addrfile.get_ipv4("test") ==
            (ipaddress.IPv4Address('1.2.3.4'), True))
    assert (empty_addrfile.get_ipv6("test") ==
            (ipaddress.IPv6Network('1234::/64'), True))
    assert not empty_addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert not empty_addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )
    assert empty_addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('5.6.7.8')
    )
    assert empty_addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('5678::/64')
    )

    empty_addrfile.set_ipv4("test2", ipaddress.IPv4Address('1.2.3.4'))
    empty_addrfile.set_ipv6("test2", ipaddress.IPv6Network('1234::/64'))
    assert (empty_addrfile.get_ipv4("test2") ==
            (ipaddress.IPv4Address('1.2.3.4'), True))
    assert (empty_addrfile.get_ipv6("test2") ==
            (ipaddress.IPv6Network('1234::/64'), True))
    assert not empty_addrfile.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('1.2.3.4')
    )
    assert not empty_addrfile.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('1234::/64')
    )
    assert empty_addrfile.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('5.6.7.8')
    )
    assert empty_addrfile.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('5678::/64')
    )


def test_existing_addrfile(empty_addrfile):
    """Test getting addresses from existing addrfile"""
    empty_addrfile.set_ipv4("test", ipaddress.IPv4Address('1.2.3.4'))
    empty_addrfile.set_ipv6("test", ipaddress.IPv6Network('1234::/64'))
    empty_addrfile.invalidate_ipv4("test2", ipaddress.IPv4Address('5.6.7.8'))
    empty_addrfile.invalidate_ipv6("test2", ipaddress.IPv6Network('5678::/64'))
    path = empty_addrfile.path

    addrfile = ruddr.addrfile.Addrfile(path)
    assert (addrfile.get_ipv4("test") ==
            (ipaddress.IPv4Address('1.2.3.4'), True))
    assert (addrfile.get_ipv6("test") ==
            (ipaddress.IPv6Network('1234::/64'), True))
    assert not empty_addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert not empty_addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )
    assert addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('5.6.7.8')
    )
    assert addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('5678::/64')
    )
    addrfile.invalidate_ipv4("test", ipaddress.IPv4Address('2.3.4.5'))
    addrfile.invalidate_ipv6("test", ipaddress.IPv6Network('2345::/64'))
    assert (addrfile.get_ipv4("test") ==
            (ipaddress.IPv4Address('2.3.4.5'), False))
    assert (addrfile.get_ipv6("test") ==
            (ipaddress.IPv6Network('2345::/64'), False))
    assert addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('2.3.4.5')
    )
    assert addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('2345::/64')
    )

    assert (addrfile.get_ipv4("test2") ==
            (ipaddress.IPv4Address('5.6.7.8'), False))
    assert (addrfile.get_ipv6("test2") ==
            (ipaddress.IPv6Network('5678::/64'), False))
    assert addrfile.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('5.6.7.8')
    )
    assert addrfile.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('5678::/64')
    )
    assert addrfile.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('1.2.3.4')
    )
    assert addrfile.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('1234::/64')
    )
    addrfile.set_ipv4("test2", ipaddress.IPv4Address('6.7.8.9'))
    addrfile.set_ipv6("test2", ipaddress.IPv6Network('6789::/64'))
    assert (addrfile.get_ipv4("test2") ==
            (ipaddress.IPv4Address('6.7.8.9'), True))
    assert (addrfile.get_ipv6("test2") ==
            (ipaddress.IPv6Network('6789::/64'), True))
    assert not addrfile.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('6.7.8.9')
    )
    assert not addrfile.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('6789::/64')
    )


def test_only_ipv4_is_set(empty_addrfile):
    """Test getting addresses when only IPv4 is set"""
    empty_addrfile.set_ipv4("test", ipaddress.IPv4Address('1.2.3.4'))
    empty_addrfile.invalidate_ipv4("test2", ipaddress.IPv4Address('5.6.7.8'))
    path = empty_addrfile.path

    addrfile = ruddr.addrfile.Addrfile(path)
    assert (addrfile.get_ipv4("test") ==
            (ipaddress.IPv4Address('1.2.3.4'), True))
    assert addrfile.get_ipv6("test") == (None, False)
    assert not empty_addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert empty_addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )
    assert empty_addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('5.6.7.8')
    )

    assert (addrfile.get_ipv4("test2") ==
            (ipaddress.IPv4Address('5.6.7.8'), False))
    assert addrfile.get_ipv6("test2") == (None, False)
    assert empty_addrfile.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('5.6.7.8')
    )
    assert empty_addrfile.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('5678::/64')
    )
    assert empty_addrfile.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('1.2.3.4')
    )


def test_only_ipv6_is_set(empty_addrfile):
    """Test getting addresses when only IPv6 is set"""
    empty_addrfile.set_ipv6("test", ipaddress.IPv6Network('1234::/64'))
    empty_addrfile.invalidate_ipv6("test2", ipaddress.IPv6Network('5678::/64'))
    path = empty_addrfile.path

    addrfile = ruddr.addrfile.Addrfile(path)
    assert addrfile.get_ipv4("test") == (None, False)
    assert (addrfile.get_ipv6("test") ==
            (ipaddress.IPv6Network('1234::/64'), True))
    assert empty_addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert not empty_addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )
    assert empty_addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('5678::/64')
    )

    assert addrfile.get_ipv4("test2") == (None, False)
    assert (addrfile.get_ipv6("test2") ==
            (ipaddress.IPv6Network('5678::/64'), False))
    assert empty_addrfile.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('5.6.7.8')
    )
    assert empty_addrfile.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('5678::/64')
    )
    assert empty_addrfile.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('1234::/64')
    )


def test_set_ipv4_invalidate_ipv6(empty_addrfile):
    """Test getting addresses when IPv4 is set successfully and IPv6 is
    invalidated"""
    empty_addrfile.set_ipv4("test", ipaddress.IPv4Address('1.2.3.4'))
    empty_addrfile.invalidate_ipv6("test", ipaddress.IPv6Network('1234::/64'))
    path = empty_addrfile.path

    addrfile = ruddr.addrfile.Addrfile(path)
    assert (addrfile.get_ipv4("test") ==
            (ipaddress.IPv4Address('1.2.3.4'), True))
    assert (addrfile.get_ipv6("test") ==
            (ipaddress.IPv6Network('1234::/64'), False))
    assert not empty_addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert empty_addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )
    assert empty_addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('5.6.7.8')
    )


def test_set_ipv6_invalidate_ipv4(empty_addrfile):
    """Test getting addresses when IPv6 is set successfully and IPv4 is
    invalidated"""
    empty_addrfile.invalidate_ipv4("test", ipaddress.IPv4Address('1.2.3.4'))
    empty_addrfile.set_ipv6("test", ipaddress.IPv6Network('1234::/64'))
    path = empty_addrfile.path

    addrfile = ruddr.addrfile.Addrfile(path)
    assert (addrfile.get_ipv4("test") ==
            (ipaddress.IPv4Address('1.2.3.4'), False))
    assert (addrfile.get_ipv6("test") ==
            (ipaddress.IPv6Network('1234::/64'), True))
    assert empty_addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert not empty_addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )
    assert empty_addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('5678::/64')
    )


def test_read_error(mocker, tmp_path):
    """Test read error in addrfile"""
    mocker.patch('builtins.open', return_value=doubles.BrokenFile())
    addrfile = ruddr.addrfile.Addrfile(tmp_path / 'addrfile')

    assert addrfile.get_ipv4("test") == (None, False)
    assert addrfile.get_ipv6("test") == (None, False)
    assert addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )


def test_set_ipv4_write_error(mocker, tmp_path):
    """Test write errors for set_ipv4"""
    mocker.patch('builtins.open',
                 return_value=doubles.BrokenFile(write_broken=True))
    addrfile = ruddr.addrfile.Addrfile(tmp_path / 'addrfile')
    with pytest.raises(OSError):
        addrfile.set_ipv4("test", ipaddress.IPv4Address('1.2.3.4'))


def test_set_ipv6_write_error(mocker, tmp_path):
    """Test write errors for set_ipv6"""
    mocker.patch('builtins.open',
                 return_value=doubles.BrokenFile(write_broken=True))
    addrfile = ruddr.addrfile.Addrfile(tmp_path / 'addrfile')
    with pytest.raises(OSError):
        addrfile.set_ipv6("test", ipaddress.IPv6Network('1234::/64'))


def test_invalidate_ipv4_write_error(mocker, tmp_path):
    """Test write errors for invalidate_ipv4"""
    mocker.patch('builtins.open',
                 return_value=doubles.BrokenFile(write_broken=True))
    addrfile = ruddr.addrfile.Addrfile(tmp_path / 'addrfile')
    with pytest.raises(OSError):
        addrfile.invalidate_ipv4("test", ipaddress.IPv4Address('1.2.3.4'))


def test_invalidate_ipv6_write_error(mocker, tmp_path):
    """Test write errors for invalidate_ipv6"""
    mocker.patch('builtins.open',
                 return_value=doubles.BrokenFile(write_broken=True))
    addrfile = ruddr.addrfile.Addrfile(tmp_path / 'addrfile')
    with pytest.raises(OSError):
        addrfile.invalidate_ipv6("test", ipaddress.IPv6Network('1234::/64'))


@pytest.mark.parametrize('contents', [
    '',
    '["1.2.3.4"]',
    '{}',
    '1',
    'true',
    'false',
    'null',
    '"1.2.3.4"',
    'invalid json',
])
def test_not_filled_json_object(addrfile_factory, contents):
    """Test reading an addrfile that's not JSON, a JSON non-object type, or is
    an empty JSON object"""
    addrfile = addrfile_factory(contents)
    assert addrfile.get_ipv4("test") == (None, False)
    assert addrfile.get_ipv6("test") == (None, False)
    assert addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )

    addrfile.set_ipv4("test", ipaddress.IPv4Address('1.2.3.4'))
    addrfile.set_ipv6("test", ipaddress.IPv6Network('1234::/64'))
    assert (addrfile.get_ipv4("test") ==
            (ipaddress.IPv4Address('1.2.3.4'), True))
    assert (addrfile.get_ipv6("test") ==
            (ipaddress.IPv6Network('1234::/64'), True))
    assert not addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert not addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )
    assert addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('5.6.7.8')
    )
    assert addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('5678::/64')
    )

    addrfile2 = ruddr.addrfile.Addrfile(addrfile.path)
    assert (addrfile2.get_ipv4("test") ==
            (ipaddress.IPv4Address('1.2.3.4'), True))
    assert (addrfile2.get_ipv6("test") ==
            (ipaddress.IPv6Network('1234::/64'), True))
    assert not addrfile2.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert not addrfile2.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )
    assert addrfile2.needs_ipv4_update(
        "test", ipaddress.IPv4Address('5.6.7.8')
    )
    assert addrfile2.needs_ipv6_update(
        "test", ipaddress.IPv6Network('5678::/64')
    )


@pytest.mark.parametrize('contents', [
    """{
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": null,
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": true,
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": false,
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": 1,
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": "1.2.3.4",
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": ["1.2.3.4"],
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv5": ["1.2.3.4", true]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true],
            "ipv5": ["1.2.3.4", true]
        },
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {
            "ipv4": ["1234::/64", true],
            "ipv6": ["1234::/64", true]
        },
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1.2.3.4", true]
        },
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": ["1.2.3.4"]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": ["1.2.3.4", true, null]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": ["1.2.3.4", 1]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": ["1.2.3.4", null]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": ["1.2.3.4", "true"]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": ["1.2.3.4", [true]]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": ["1.2.3.4", {}]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": [["1.2.3.4"], true]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": [{}, true]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": [1234, true]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": [true, true]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": ["1.2.3.4.5", true]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": [["1.2.3.4", true]]},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": {"1.2.3.4": true}},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": "1.2.3.4"},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": null},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": true},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": false},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
    """{
        "test": {"ipv4": 1234},
        "test2": {
            "ipv4": ["1.2.3.4", true],
            "ipv6": ["1234::/64", true]
        }
    }""",
])
def test_one_invalid_key(addrfile_factory, contents):
    """Test reading an addrfile with one malformed/missing key and one valid
    key"""
    addrfile = addrfile_factory(contents)

    assert addrfile.get_ipv4("test") == (None, False)
    assert addrfile.get_ipv6("test") == (None, False)
    assert addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )

    assert (addrfile.get_ipv4("test2") ==
            (ipaddress.IPv4Address('1.2.3.4'), True))
    assert (addrfile.get_ipv6("test2") ==
            (ipaddress.IPv6Network('1234::/64'), True))
    assert not addrfile.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('1.2.3.4')
    )
    assert not addrfile.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('1234::/64')
    )
    assert addrfile.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('5.6.7.8')
    )
    assert addrfile.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('5678::/64')
    )

    addrfile.set_ipv4("test", ipaddress.IPv4Address('1.2.3.4'))
    addrfile.set_ipv6("test", ipaddress.IPv6Network('1234::/64'))
    assert (addrfile.get_ipv4("test") ==
            (ipaddress.IPv4Address('1.2.3.4'), True))
    assert (addrfile.get_ipv6("test") ==
            (ipaddress.IPv6Network('1234::/64'), True))
    assert not addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert not addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )
    assert addrfile.needs_ipv4_update(
        "test", ipaddress.IPv4Address('5.6.7.8')
    )
    assert addrfile.needs_ipv6_update(
        "test", ipaddress.IPv6Network('5678::/64')
    )

    addrfile2 = ruddr.addrfile.Addrfile(addrfile.path)

    assert (addrfile2.get_ipv4("test") ==
            (ipaddress.IPv4Address('1.2.3.4'), True))
    assert (addrfile2.get_ipv6("test") ==
            (ipaddress.IPv6Network('1234::/64'), True))
    assert not addrfile2.needs_ipv4_update(
        "test", ipaddress.IPv4Address('1.2.3.4')
    )
    assert not addrfile2.needs_ipv6_update(
        "test", ipaddress.IPv6Network('1234::/64')
    )
    assert addrfile2.needs_ipv4_update(
        "test", ipaddress.IPv4Address('5.6.7.8')
    )
    assert addrfile2.needs_ipv6_update(
        "test", ipaddress.IPv6Network('5678::/64')
    )

    assert (addrfile2.get_ipv4("test2") ==
            (ipaddress.IPv4Address('1.2.3.4'), True))
    assert (addrfile2.get_ipv6("test2") ==
            (ipaddress.IPv6Network('1234::/64'), True))
    assert not addrfile2.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('1.2.3.4')
    )
    assert not addrfile2.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('1234::/64')
    )
    assert addrfile2.needs_ipv4_update(
        "test2", ipaddress.IPv4Address('5.6.7.8')
    )
    assert addrfile2.needs_ipv6_update(
        "test2", ipaddress.IPv6Network('5678::/64')
    )



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Built in updaters and the updater base class"""

from .updater import (BaseUpdater, Updater, OneWayUpdater, TwoWayUpdater,
                      TwoWayZoneUpdater)

from . import duckdns
from . import freedns
from . import gandi
from . import he
from . import standard

updaters = {
    'duckdns': duckdns.DuckDNSUpdater,
    'freedns': freedns.FreeDNSUpdater,
    'gandi': gandi.GandiUpdater,
    'he': he.HEUpdater,
    'standard': standard.StandardUpdater,
}

__all__ = ['BaseUpdater', 'Updater', 'OneWayUpdater', 'TwoWayUpdater',
           'TwoWayZoneUpdater']



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test doubles for use in test classes and fixtures"""
import collections
import errno
import ipaddress
import itertools
from typing import List, Tuple, Set, Optional, Dict

import ruddr
import ruddr.util
from ruddr import NotifierSetupError, Addrfile, PublishError


def raise_or_return(result):
    if isinstance(result, BaseException):
        raise result
    if isinstance(result, type) and issubclass(result, BaseException):
        raise result
    return result


class BrokenFile:
    """File-like object that raises an exception when being read from"""
    def __init__(self, write_broken=False):
        self.write_broken = write_broken

    def __iter__(self):
        if self.write_broken:
            return iter([])
        else:
            raise OSError(errno.ETIMEDOUT, "timeout")

    def read(self, *_):
        if self.write_broken:
            return ""
        else:
            raise OSError(errno.ETIMEDOUT, "timeout")

    def write(self, *_):
        if self.write_broken:
            raise OSError(errno.ETIMEDOUT, "timeout")
        else:
            return 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class MockZoneSplitter(ruddr.util.ZoneSplitter):
    """A mock ZoneSplitter that tracks the domains that have been split"""

    split_domains: Optional[List[str]] = []

    def __init__(self, datadir: str):
        super().__init__(datadir)
        #: List of domains that have been split by this instance

    def split(self, domain: str) -> Tuple[str, str]:
        self.split_domains.append(domain)
        return super().split(domain)

    @classmethod
    def clear_domains(cls):
        cls.split_domains = []


class FakeNotifier(ruddr.BaseNotifier):
    """A simple notifier that notifies on demand. Extends BaseNotifier."""
    # Note: Tests can trigger notifying by calling .notify_ipv4() and
    # .notify_ipv6() directly

    def __init__(self, name, config):
        super().__init__(name, config)
        self.config = config
        # Config vars to test .ipv4_ready() and .ipv6_ready()
        self._ipv4_ready = (config.get('ipv4_ready', 'true').lower() in
                            ('true', 'yes', 'on', '1'))
        self._ipv6_ready = (config.get('ipv6_ready', 'true').lower() in
                            ('true', 'yes', 'on', '1'))

    def ipv4_ready(self):
        return self._ipv4_ready

    def ipv6_ready(self):
        return self._ipv6_ready


class MockNotifier(ruddr.Notifier):
    """A mock Notifier whose checks succeed and fail in a predetermined order
    and which tracks when calls were made"""

    def __init__(self, name, config, success_sequence=None,
                 setup_implemented=True, teardown_implemented=True,
                 check_implemented=True, setup_error=False):
        super().__init__(name, config)

        #: The order of successes and fails for check_once
        if success_sequence is None:
            self.success_iter = itertools.repeat(True)
        else:
            self.success_iter = iter(success_sequence)

        self.setup_implemented = setup_implemented
        self.teardown_implemented = teardown_implemented
        self.check_implemented = check_implemented
        self.setup_error = setup_error

        #: The order the abstract methods were called
        self.call_sequence = []

        # Used only for test_manager.py
        self.stop_count = 0

    @property
    def setup_count(self):
        return sum(call == 'setup' for call in self.call_sequence)

    @property
    def teardown_count(self):
        return sum(call == 'teardown' for call in self.call_sequence)

    @property
    def check_count(self):
        return sum(call == 'check' for call in self.call_sequence)

    def setup(self):
        self.call_sequence.append('setup')
        if not self.setup_implemented:
            raise NotImplementedError
        if self.setup_error:
            raise NotifierSetupError

    def teardown(self):
        self.call_sequence.append('teardown')
        if not self.teardown_implemented:
            raise NotImplementedError

    def check_once(self):
        self.call_sequence.append('check')
        if not self.check_implemented:
            raise NotImplementedError
        success = next(self.success_iter)
        if not success:
            raise ruddr.NotifyError

    # test_manager needs to tell if stop was called even if start was not
    def stop(self) -> None:
        self.stop_count += 1
        super().stop()

    def join_first_check(self):
        """Wait for first check after :meth:`start` to finish"""
        self.first_check.join()


class MockBaseUpdater(ruddr.BaseUpdater):
    """Simple mock updater that keeps a list of IP updates it receives"""

    def __init__(self, name, addrfile=None, config=None, err_sequence=None):
        super().__init__(name, addrfile)
        self.config = config
        self.published_addresses = []
        self.initial_update_calls: List[Tuple[bool, bool]] = []

        #: The order of successes and fails for retry_test. None means success,
        #: an error means raise that error
        if err_sequence is None:
            self.err_iter = itertools.repeat(None)
        else:
            self.err_iter = iter(err_sequence)
        #: The list of calls to retry_test
        self.retry_sequence = []

    @ruddr.updaters.updater.Retry
    def retry_test(self, param: int):
        """A dummy function to test @Retry. Only used in test_baseupdater."""
        self.retry_sequence.append(param)
        error = next(self.err_iter)
        if error is not None:
            raise error

    def initial_update(self, ipv4_attached: bool, ipv6_attached: bool):
        self.initial_update_calls.append((ipv4_attached, ipv6_attached))

    def update_ipv4(self, address):
        self.published_addresses.append(address)

    def update_ipv6(self, network):
        self.published_addresses.append(network)


class MockUpdater(ruddr.Updater):
    """Mock Updater that tracks calls to its abstract functions"""

    def __init__(self, name: str, addrfile: Addrfile,
                 ipv4_errors=None, ipv6_errors=None,
                 ipv4_implemented=True, ipv6_implemented=True):
        super().__init__(name, addrfile)
        self.ipv4s_published: List[ipaddress.IPv4Address] = []
        self.ipv6s_published: List[ipaddress.IPv6Network] = []

        if ipv4_errors is None:
            self.ipv4_iter = itertools.repeat(None)
        else:
            self.ipv4_iter = iter(ipv4_errors)
        if ipv6_errors is None:
            self.ipv6_iter = itertools.repeat(None)
        else:
            self.ipv6_iter = iter(ipv6_errors)

        self.ipv4_implemented = ipv4_implemented
        self.ipv6_implemented = ipv6_implemented

    def publish_ipv4(self, address: ipaddress.IPv4Address):
        self.ipv4s_published.append(address)
        if not self.ipv4_implemented:
            raise NotImplementedError
        error = next(self.ipv4_iter)
        if error is not None:
            raise error

    def publish_ipv6(self, network: ipaddress.IPv6Network):
        self.ipv6s_published.append(network)
        if not self.ipv6_implemented:
            raise NotImplementedError
        error = next(self.ipv6_iter)
        if error is not None:
            raise error


class MockOneWayUpdater(ruddr.OneWayUpdater):
    """Mock OneWayUpdater that tracks calls to its abstract functions"""

    def __init__(self, name: str, addrfile: Addrfile,
                 ipv4_implemented=True, ipv6_implemented=True,
                 ipv4_errors=None, ipv6_errors=None):
        super().__init__(name, addrfile)
        self.ipv4s_published: collections.Counter[
            Tuple[str, ipaddress.IPv4Address]
        ] = collections.Counter()
        self.ipv6s_published: collections.Counter[
            Tuple[str, ipaddress.IPv6Address]
        ] = collections.Counter()

        self.ipv4_implemented = ipv4_implemented
        self.ipv6_implemented = ipv6_implemented

        # Sets of hostnames which should raise PublishError when published
        self.ipv4_errors: Set[str] = set()
        if ipv4_errors is not None:
            for host in ipv4_errors:
                self.ipv4_errors.add(host)
        self.ipv6_errors: Set[str] = set()
        if ipv6_errors is not None:
            for host in ipv6_errors:
                self.ipv6_errors.add(host)

    def publish_ipv4_one_host(self,
                              hostname: str,
                              address: ipaddress.IPv4Address):
        self.ipv4s_published[(hostname, address)] += 1
        if not self.ipv4_implemented:
            raise NotImplementedError
        if hostname in self.ipv4_errors:
            raise PublishError

    def publish_ipv6_one_host(self,
                              hostname: str,
                              address: ipaddress.IPv6Address):
        self.ipv6s_published[(hostname, address)] += 1
        if not self.ipv6_implemented:
            raise NotImplementedError
        if hostname in self.ipv6_errors:
            raise PublishError


class MockTwoWayZoneUpdater(ruddr.TwoWayZoneUpdater):
    """Mock TwoWayZoneUpdater that tracks calls to its abstract methods"""

    def __init__(self, name: str, addrfile: Addrfile, datadir: str,
                 get_zones_result=None,
                 fetch_zone_ipv4s_result=None,
                 fetch_zone_ipv6s_result=None,
                 fetch_subdomain_ipv4s_result=None,
                 fetch_subdomain_ipv6s_result=None,
                 put_zone_ipv4s_result=None,
                 put_zone_ipv6s_result=None,
                 put_subdomain_ipv4_result=None,
                 put_subdomain_ipv6s_result=None):
        super().__init__(name, addrfile, datadir)

        self.get_zones_result = get_zones_result
        self.fetch_zone_ipv4s_result = fetch_zone_ipv4s_result
        self.fetch_zone_ipv6s_result = fetch_zone_ipv6s_result
        self.fetch_subdomain_ipv4s_result = fetch_subdomain_ipv4s_result
        self.fetch_subdomain_ipv6s_result = fetch_subdomain_ipv6s_result
        self.put_zone_ipv4s_result = put_zone_ipv4s_result
        self.put_zone_ipv6s_result = put_zone_ipv6s_result
        self.put_subdomain_ipv4_result = put_subdomain_ipv4_result
        self.put_subdomain_ipv6s_result = put_subdomain_ipv6s_result

        self.get_zones_call_count = 0
        self.fetch_zone_ipv4s_calls = []
        self.fetch_zone_ipv6s_calls = []
        self.fetch_subdomain_ipv4s_calls = []
        self.fetch_subdomain_ipv6s_calls = []
        self.put_zone_ipv4s_calls = []
        self.put_zone_ipv6s_calls = []
        self.put_subdomain_ipv4_calls = []
        self.put_subdomain_ipv6s_calls = []

    def get_zones(self) -> List[str]:
        self.get_zones_call_count += 1
        if self.get_zones_result is None:
            raise NotImplementedError
        return raise_or_return(self.get_zones_result)

    def fetch_zone_ipv4s(
        self, zone: str
    ) -> List[Tuple[str, ipaddress.IPv4Address, Optional[int]]]:
        self.fetch_zone_ipv4s_calls.append(zone)
        if self.fetch_zone_ipv4s_result is None:
            raise NotImplementedError
        result = self.fetch_zone_ipv4s_result[zone]
        return raise_or_return(result)

    def fetch_zone_ipv6s(
        self, zone: str
    ) -> List[Tuple[str, ipaddress.IPv6Address, Optional[int]]]:
        self.fetch_zone_ipv6s_calls.append(zone)
        if self.fetch_zone_ipv6s_result is None:
            raise NotImplementedError
        result = self.fetch_zone_ipv6s_result[zone]
        return raise_or_return(result)

    def fetch_subdomain_ipv4s(
        self, subdomain: str, zone: str
    ) -> List[Tuple[ipaddress.IPv4Address, Optional[int]]]:
        self.fetch_subdomain_ipv4s_calls.append((subdomain, zone))
        if self.fetch_subdomain_ipv4s_result is None:
            raise NotImplementedError
        result = self.fetch_subdomain_ipv4s_result[(subdomain, zone)]
        return raise_or_return(result)

    def fetch_subdomain_ipv6s(
        self, subdomain: str, zone: str
    ) -> List[Tuple[ipaddress.IPv6Address, Optional[int]]]:
        self.fetch_subdomain_ipv6s_calls.append((subdomain, zone))
        if self.fetch_subdomain_ipv6s_result is None:
            raise NotImplementedError
        result = self.fetch_subdomain_ipv6s_result[(subdomain, zone)]
        return raise_or_return(result)

    def put_zone_ipv4s(
        self,
        zone: str,
        records: Dict[str, Tuple[List[ipaddress.IPv4Address], Optional[int]]]
    ):
        self.put_zone_ipv4s_calls.append((zone, records))
        if self.put_zone_ipv4s_result is None:
            raise NotImplementedError
        result = self.put_zone_ipv4s_result[zone]
        return raise_or_return(result)

    def put_zone_ipv6s(
        self,
        zone: str,
        records: Dict[str, Tuple[List[ipaddress.IPv6Address], Optional[int]]]
    ):
        self.put_zone_ipv6s_calls.append((zone, records))
        if self.put_zone_ipv6s_result is None:
            raise NotImplementedError
        result = self.put_zone_ipv6s_result[zone]
        return raise_or_return(result)

    def put_subdomain_ipv4(self, subdomain: str, zone: str,
                           address: ipaddress.IPv4Address, ttl: Optional[int]):
        self.put_subdomain_ipv4_calls.append((subdomain, zone, address, ttl))
        if self.put_subdomain_ipv4_result is None:
            raise NotImplementedError
        result = self.put_subdomain_ipv4_result[(subdomain, zone)]
        return raise_or_return(result)

    def put_subdomain_ipv6s(self, subdomain: str, zone: str,
                            addresses: List[ipaddress.IPv6Address],
                            ttl: Optional[int]):
        self.put_subdomain_ipv6s_calls.append((subdomain, zone,
                                               addresses, ttl))
        if self.put_subdomain_ipv6s_result is None:
            raise NotImplementedError
        result = self.put_subdomain_ipv6s_result[(subdomain, zone)]
        return raise_or_return(result)


class MockTwoWayUpdater(ruddr.TwoWayUpdater):
    """Mock TwoWayUpdater that tracks calls to its abstract methods"""

    def __init__(self, name: str, addrfile: Addrfile, datadir: str,
                 fetch_all_ipv4s_result=None,
                 fetch_all_ipv6s_result=None,
                 fetch_domain_ipv4s_result=None,
                 fetch_domain_ipv6s_result=None,
                 put_all_ipv4s_result=None,
                 put_all_ipv6s_result=None,
                 put_domain_ipv4_result=None,
                 put_domain_ipv6s_result=None):
        super().__init__(name, addrfile, datadir)

        self.fetch_all_ipv4s_result = fetch_all_ipv4s_result
        self.fetch_all_ipv6s_result = fetch_all_ipv6s_result
        self.fetch_domain_ipv4s_result = fetch_domain_ipv4s_result
        self.fetch_domain_ipv6s_result = fetch_domain_ipv6s_result
        self.put_all_ipv4s_result = put_all_ipv4s_result
        self.put_all_ipv6s_result = put_all_ipv6s_result
        self.put_domain_ipv4_result = put_domain_ipv4_result
        self.put_domain_ipv6s_result = put_domain_ipv6s_result

        self.fetch_all_ipv4s_count = 0
        self.fetch_all_ipv6s_count = 0
        self.fetch_domain_ipv4s_calls = []
        self.fetch_domain_ipv6s_calls = []
        self.put_all_ipv4s_calls = []
        self.put_all_ipv6s_calls = []
        self.put_domain_ipv4_calls = []
        self.put_domain_ipv6s_calls = []

    def fetch_all_ipv4s(
        self
    ) -> List[Tuple[str, ipaddress.IPv4Address, Optional[int]]]:
        self.fetch_all_ipv4s_count += 1
        if self.fetch_all_ipv4s_result is None:
            raise NotImplementedError
        return raise_or_return(self.fetch_all_ipv4s_result)

    def fetch_all_ipv6s(
        self
    ) -> List[Tuple[str, ipaddress.IPv6Address, Optional[int]]]:
        self.fetch_all_ipv6s_count += 1
        if self.fetch_all_ipv6s_result is None:
            raise NotImplementedError
        return raise_or_return(self.fetch_all_ipv6s_result)

    def fetch_domain_ipv4s(
        self, domain: str
    ) -> List[Tuple[ipaddress.IPv4Address, Optional[int]]]:
        self.fetch_domain_ipv4s_calls.append(domain)
        if self.fetch_domain_ipv4s_result is None:
            raise NotImplementedError
        result = self.fetch_domain_ipv4s_result[domain]
        return raise_or_return(result)

    def fetch_domain_ipv6s(
        self, domain: str
    ) -> List[Tuple[ipaddress.IPv6Address, Optional[int]]]:
        self.fetch_domain_ipv6s_calls.append(domain)
        if self.fetch_domain_ipv6s_result is None:
            raise NotImplementedError
        result = self.fetch_domain_ipv6s_result[domain]
        return raise_or_return(result)

    def put_all_ipv4s(
        self,
        records: Dict[str, Tuple[List[ipaddress.IPv4Address], Optional[int]]]
    ):
        self.put_all_ipv4s_calls.append(records)
        if self.put_all_ipv4s_result is None:
            raise NotImplementedError
        return raise_or_return(self.put_all_ipv4s_result)

    def put_all_ipv6s(
        self,
        records: Dict[str, Tuple[List[ipaddress.IPv6Address], Optional[int]]]
    ):
        self.put_all_ipv6s_calls.append(records)
        if self.put_all_ipv6s_result is None:
            raise NotImplementedError
        return raise_or_return(self.put_all_ipv6s_result)

    def put_domain_ipv4(self, domain: str,
                        address: ipaddress.IPv4Address,
                        ttl: Optional[int]):
        self.put_domain_ipv4_calls.append((domain, address, ttl))
        if self.put_domain_ipv4_result is None:
            raise NotImplementedError
        result = self.put_domain_ipv4_result[domain]
        return raise_or_return(result)

    def put_domain_ipv6s(self, domain: str,
                         addresses: List[ipaddress.IPv6Address],
                         ttl: Optional[int]):
        self.put_domain_ipv6s_calls.append((domain, addresses, ttl))
        if self.put_domain_ipv6s_result is None:
            raise NotImplementedError
        result = self.put_domain_ipv6s_result[domain]
        return raise_or_return(result)



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import ipaddress
import logging

import pytest

import ruddr
from ruddr import FatalPublishError, PublishError


def test_member_vars(updater_factory, empty_addrfile):
    """Test BaseUpdater has basic member variables name, log, and addrfile"""
    updater = updater_factory(addrfile=empty_addrfile)
    assert updater.name == "mock_updater_1"
    assert isinstance(updater.log, logging.Logger)
    assert updater.log.name == f"ruddr.updater.{updater.name}"
    assert updater.addrfile is empty_addrfile


@pytest.mark.parametrize('sequence', [
    # (Error|None, 'new'|'repeat'|'retry', delay until next)
    # Error ignored for repeat calls
    # Last delay must extend until any pending retries expire

    # 0: Single successful call
    [(None, 'new', 1)],
    # 1: Two successful calls
    [(None, 'new', 1),
     (None, 'new', 0)],
    # 2: Fail, retry succeeds
    [(PublishError, 'new', 300),
     (None, 'retry', 0)],
    # 3: Fail, new call before retry
    [(PublishError, 'new', 100),
     (None, 'new', 200)],
    # 4: Fail, retry fail, success
    [(PublishError, 'new', 300),
     (PublishError, 'retry', 600),
     (None, 'retry', 0)],
    # 5: Fail, retry fail, repeat call before retry, retry happens as scheduled
    [(PublishError, 'new', 300),
     (PublishError, 'retry', 100),
     (None, 'repeat', 500),
     (None, 'retry', 0)],
    # 6: Fail, retry fail, success, fail, retry after min delay
    [(PublishError, 'new', 300),
     (PublishError, 'retry', 600),
     (None, 'retry', 0),
     (PublishError, 'new', 300),
     (None, 'retry', 0)],
    # 7: Fail, retry fail, retry fail, new call fail, retry after min delay
    [(PublishError, 'new', 300),
     (PublishError, 'retry', 600),
     (PublishError, 'retry', 1100),
     (PublishError, 'new', 300),
     (None, 'retry', 0)],
    # 8: Fail, retry fail until max delay hit, retry fail one more time,
    #    success
    [(PublishError, 'new', 300),
     (PublishError, 'retry', 600),
     (PublishError, 'retry', 1200),
     (PublishError, 'retry', 2400),
     (PublishError, 'retry', 4800),
     (PublishError, 'retry', 9600),
     (PublishError, 'retry', 19200),
     (PublishError, 'retry', 38400),
     (PublishError, 'retry', 76800),
     (PublishError, 'retry', 86400),
     (PublishError, 'retry', 86400),
     (None, 'retry', 0)],
    # 9: Fail, retry fail until max delay hit, retry fail one more time, new
    #    call fail, retry after min delay
    [(PublishError, 'new', 300),
     (PublishError, 'retry', 600),
     (PublishError, 'retry', 1200),
     (PublishError, 'retry', 2400),
     (PublishError, 'retry', 4800),
     (PublishError, 'retry', 9600),
     (PublishError, 'retry', 19200),
     (PublishError, 'retry', 38400),
     (PublishError, 'retry', 76800),
     (PublishError, 'retry', 86400),
     (PublishError, 'retry', 86300),
     (PublishError, 'new', 300),
     (None, 'retry', 0)],
    # 10: Fatal right away
    [(FatalPublishError, 'new', 0)],
    # 11: Success, fatal
    [(None, 'new', 1),
     (FatalPublishError, 'new', 0)],
    # 12: Fail, retry fail, retry fail, fatal
    [(PublishError, 'new', 300),
     (PublishError, 'retry', 600),
     (PublishError, 'retry', 1200),
     (FatalPublishError, 'retry', 0)],
    # 13: Fail, retry fail, retry fail, new call fatal
    [(PublishError, 'new', 300),
     (PublishError, 'retry', 600),
     (PublishError, 'retry', 1100),
     (FatalPublishError, 'new', 100)],
    # 14: Fail, retry fail until max delay hit, retry fail one more time, fatal
    [(PublishError, 'new', 300),
     (PublishError, 'retry', 600),
     (PublishError, 'retry', 1200),
     (PublishError, 'retry', 2400),
     (PublishError, 'retry', 4800),
     (PublishError, 'retry', 9600),
     (PublishError, 'retry', 19200),
     (PublishError, 'retry', 38400),
     (PublishError, 'retry', 76800),
     (PublishError, 'retry', 86400),
     (PublishError, 'retry', 86400),
     (FatalPublishError, 'retry', 0)],
])
def test_retry(sequence, updater_factory, advance):
    err_sequence = [x[0] for x in sequence]
    updater = updater_factory(err_sequence=err_sequence)

    # Check retry sequence
    arg = 0
    expected_seq = []
    expect_retry = False
    halted = False
    for err, kind, delay in sequence:
        # kind 'new' = new call with new param
        # kind 'repeat' = new call with same param (ignored)
        # kind 'retry' = automatic retry
        if kind == 'new':
            arg += 1
            updater.retry_test(arg)
            expected_seq.append(arg)
        elif kind == 'repeat':
            updater.retry_test(arg)
        elif expect_retry:
            expected_seq.append(arg)

        assert updater.retry_sequence == expected_seq

        if kind != 'repeat':
            expect_retry = err is not None
        advance.by(delay)
        if isinstance(err, FatalPublishError):
            halted = True

    if halted:
        assert updater.halt
    assert updater.retry_sequence == expected_seq
    # No more retries
    assert advance.count_running() == 0


@pytest.mark.parametrize(('network', 'address', 'expected'), [
    (ipaddress.IPv6Network('::/128'),
     ipaddress.IPv6Address('1:1:1:1:1:1:1:1'),
     ipaddress.IPv6Address('::')),
    (ipaddress.IPv6Network('::2/128'),
     ipaddress.IPv6Address('1:1:1:1:1:1:1:1'),
     ipaddress.IPv6Address('::2')),
    (ipaddress.IPv6Network('ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff/128'),
     ipaddress.IPv6Address('1:1:1:1:1:1:1:1'),
     ipaddress.IPv6Address('ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff')),
    (ipaddress.IPv6Network('2222::/64'),
     ipaddress.IPv6Address('1:1:1:1:1:1:1:1'),
     ipaddress.IPv6Address('2222::1:1:1:1')),
    (ipaddress.IPv6Network('::/64'),
     ipaddress.IPv6Address('1:1:1:1:1:1:1:1'),
     ipaddress.IPv6Address('::1:1:1:1')),
    (ipaddress.IPv6Network('ffff:ffff:ffff:ffff::/64'),
     ipaddress.IPv6Address('1:1:1:1:1:1:1:1'),
     ipaddress.IPv6Address('ffff:ffff:ffff:ffff:1:1:1:1')),
    (ipaddress.IPv6Network('::/0'),
     ipaddress.IPv6Address('1:1:1:1:1:1:1:1'),
     ipaddress.IPv6Address('1:1:1:1:1:1:1:1')),
])
def test_replace_ipv6_prefix(network, address, expected):
    assert ruddr.updaters.BaseUpdater.replace_ipv6_prefix(
        network, address
    ) == expected



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import logging
import logging.handlers
import signal
import sys

from . import configuration, manager
from .exceptions import ConfigError, RuddrSetupError
from .util import sdnotify


def parse_args(argv):
    """Parse command line arguments

    :param argv: Either ``None`` or a list of arguments
    :returns: a :class:`argparse.Namespace` containing the parsed arguments
    """
    parser = argparse.ArgumentParser(
        description="Robotic Updater for Dynamic DNS Records",
        epilog="SIGUSR1 will cause a running instance to immediately check and"
               " update the current IP address(es) if possible",
    )
    parser.add_argument("-c", "--configfile", default="/etc/ruddr.conf",
                        help="Path to the config file")
    parser.add_argument("-d", "--debug-logs", action="store_true",
                        help="Increase verbosity of logging significantly")
    parser.add_argument("-s", "--stderr", action="store_true",
                        help="Log to stderr instead of syslog or file")
    return parser.parse_args(argv)


def main(argv=None):
    """Main entry point when run as a standalone program

    :param argv: List of arguments. If ``None``, read :data:`sys.argv`.
    """
    args = parse_args(argv)
    try:
        conf = configuration.read_config_from_path(args.configfile)
    except ConfigError as e:
        print("Config error:", e, file=sys.stderr)
        sys.exit(2)

    # Set up logging handler
    if args.stderr:
        logfile = 'stderr'
    else:
        logfile = conf.main.get('log', 'syslog')
    if logfile == 'syslog':
        log_handler = logging.handlers.SysLogHandler()
    elif logfile == 'stderr':
        log_handler = logging.StreamHandler()
    else:
        log_handler = logging.FileHandler(logfile)
    log = logging.getLogger('ruddr')
    log.addHandler(log_handler)
    if args.debug_logs:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)

    # Start up the actual DDNS code
    try:
        ddns_manager = manager.DDNSManager(conf)
        ddns_manager.start()
    except RuddrSetupError:
        log.critical("Ruddr failed to start.")
        sys.exit(1)

    # Notify systemd, if applicable
    sdnotify.ready()

    # Do an immediate update on SIGUSR1
    def handle_sigusr1(sig, _):
        log.info("Received signal: %s", signal.Signals(sig).name)
        ddns_manager.do_notify()
    signal.signal(signal.SIGUSR1, handle_sigusr1)

    # Wait for SIGINT (^C) or SIGTERM
    def handle_signals(sig, _):
        log.info("Received signal: %s", signal.Signals(sig).name)
        sdnotify.stopping()
        ddns_manager.stop()
    signal.signal(signal.SIGINT, handle_signals)
    signal.signal(signal.SIGTERM, handle_signals)



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Tools for splitting a domain into subdomain part and zone part"""

import os
import os.path
from typing import Tuple, Optional
import sys
if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

import tldextract


class _TLDExtractor(Protocol):
    def __call__(
        self, url: str, include_psl_private_domains: Optional[bool] = ...
    ) -> tldextract.tldextract.ExtractResult:
        ...


class ZoneSplitter:
    """A utility to split domains into subdomain part and zone part using the
    `Public Suffix List`_

    .. _Public Suffix List: https://publicsuffix.org/

    :param datadir: The data directory configured for Ruddr
    :raises OSError: if the public suffix list could not be fetched due to I/O
                     issues (cache directory could not be created, could not
                     connect to server, etc.)
    """

    def __init__(self, datadir: str):
        self._tld_cache_dir: str = os.path.join(datadir, 'tldextract')
        os.makedirs(os.path.dirname(self._tld_cache_dir), exist_ok=True)
        self._extract_func: _TLDExtractor = tldextract.TLDExtract(
            cache_dir=self._tld_cache_dir,
            include_psl_private_domains=True,
        )

    def split(self, domain: str) -> Tuple[str, str]:
        """Split a domain name into subdomain part and zone part

        :param domain: The FQDN to split
        :return: A tuple with the two parts. The subdomain part may be empty if
                 the FQDN was the root domain of its zone.
        """
        subdomain, domain, suffix = self._extract_func(domain)
        zone = '.'.join(part for part in (domain, suffix) if part != '')
        return (subdomain, zone)



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Ruddr, the Robotic Updater for Dynamic DNS Records

Top-level module, containing classes and objects useful to custom notifiers and
updaters.
"""

from .addrfile import Addrfile
from .configuration import Config, read_config, read_config_from_path
from .exceptions import (RuddrException, RuddrSetupError, ConfigError,
                         NotifierSetupError, NotStartedError, NotifyError,
                         PublishError, FatalPublishError)
from .manager import DDNSManager
from .notifiers import BaseNotifier, Notifier
from .updaters import (Updater, BaseUpdater, OneWayUpdater, TwoWayUpdater,
                       TwoWayZoneUpdater)
